# Copyright {{ year }} {{ author }}
# See LICENSE file for licensing details.

# For a complete list of supported options, see:
# https://discourse.charmhub.io/t/charm-metadata-v2/3674/15
name: {{ name }}
display-name: |
  TEMPLATE-TODO: fill out a display name for the Charmcraft store
description: |
  TEMPLATE-TODO: fill out the charm's description
summary: |
  TEMPLATE-TODO: fill out the charm's summary

# TEMPLATE-TODO: replace with containers for your workload (delete for non-k8s)
containers:
  httpbin:
    resource: httpbin-image

# TEMPLATE-TODO: each container defined above must specify an oci-image resource
resources:
  httpbin-image:
    type: oci-image
    description: OCI image for httpbin (kennethreitz/httpbin)
