# Copyright {{ year }} {{ author }}
# See LICENSE file for licensing details.
#
# TEMPLATE-TODO: change this example to suit your needs.
# If you don't need a config, you can remove the file entirely.
# It ties in to the example _on_config_changed handler in src/charm.py
#
# Learn more about config at: https://juju.is/docs/sdk/config

options:
  thing:
    default: 🎁
    description: A thing used by the charm.
    type: string
